"""General cymetric tools.
"""
from __future__ import unicode_literals, print_function
import os
import sys
import warnings

import numpy as np
import pandas as pd

try:
    from pyne import data
    import pyne.enrichment as enr
    from pyne import nucname
    HAVE_PYNE = True
except ImportError:
    HAVE_PYNE = False

from cyclus import lib


EXT_BACKENDS = {'.h5': lib.Hdf5Back, '.sqlite': lib.SqliteBack}


def dbopen(fname):
    """Opens a Cyclus database."""
    _, ext = os.path.splitext(fname)
    if ext not in EXT_BACKENDS:
        msg = ('The backend database type of {0!r} could not be determined from '
               'extension {1!r}.')
        raise ValueError(msg.format(fname, ext))
    db = EXT_BACKENDS[ext](fname)
    return db


def raw_to_series(df, idx, val):
    """Convert data frame to series with multi-index."""
    d = df.set_index(list(map(str, idx)))
    s = df[val]
    s.index = d.index
    return s


def merge_and_fillna_col(left, right, lcol, rcol, how='inner', on=None):
    """Merges two dataframes and fills the values of the left column
    with the values from the right column. A copy of left is returned.

    Parameters
    ----------
    left : pd.DataFrame
        The left data frame
    right : pd.DataFrame
        The right data frame
    lcol : str
        The left column name
    rcol : str
        The right column name
    how : str, optional
        How to perform merge, same as in pd.merge()
    on : list of str, optional
        Which columns to merge on, same as in pd.merge()
    """
    m = pd.merge(left, right, how=how, on=on)
    f = m[lcol].fillna(m[rcol])
    left[lcol] = f
    return left


def ensure_dt_bytes(dt):
    """Ensures that a structured numpy dtype is given in a Python 2 & 3
    compatible way.
    """
    if sys.version_info[0] > 2:
        return dt
    dety = []
    for t in dt:
        t0 = t[0].encode() if isinstance(t[0], unicode) else t[0]
        t1 = t[1].encode() if isinstance(t[1], unicode) else t[1]
        ty = (t0, t1)
        if len(t) == 3:
            ty = ty + t[2:]
        dety.append(ty)
    return dety


def format_nucs(nucs):
    """
    format the nuclide  provided by the users into a standard format:
    ZZAASSSS.

    Parameters
    ----------
    nucs :  of nuclides
    """
    raise_no_pyne('Unable to format nuclide !', HAVE_PYNE)

    return [nucname.id(nuc) for nuc in nucs]


def add_missing_time_step(df, ref_time):
    """
    Add the missing time step to a Panda Data Frame.

    Parameters
    ----------
    df : Pandas Data Frame
    ref_time :  of the time step references (Coming from TimeStep metrics)
    """
    ref_time.rename(index=str, columns={'TimeStep': 'Time'}, inplace=True)

    if 'SimId' in ref_time.columns.values:
        ref_time.drop('SimId', 1, inplace=True)
    df = pd.merge(ref_time, df, how="outer")
    df.fillna(0, inplace=True)
    return df


def merge(df, base_col, add_df, add_col):
    """
    Merge some additionnal columns fram an additionnal Pandas Data Frame
    onother one and then remove the second base column (keeping SimID
    information).

    Parameters
    ----------
    df: Pandas Data Frame
    base_col:  of the base columns names
    add_df: Pandas Data Frame to add in the df one
    add_col: columns to be added
    """
    df = pd.merge(add_df[add_col], df, on=base_col)
    df.drop(base_col[1], 1)
    return df


def reduce(df, rdc_):
    """
    Filter the df Pandas Data Frame according to the rdc_ (list of item
    in the corresponding columns).

    Parameters
    ----------
    df: Pandas Data Frame
    rdc_: list of pair of string and string list.
    """
    for rdc in rdc_:
        if len(rdc[1]) != 0:
            df = df[df[rdc[0]].isin(rdc[1])]
        else:
            wng_msg = "Empty  provided for " + rdc[0] + " key."
            warnings.warn(wng_msg, UserWarning)
    return df


def raise_no_pyne(msg, have_pyne=False):
    """Raise an error when PyNE cannot be found."""
    if not have_pyne:
        raise ImportError('pyne could not be imported: ' + msg)


def raise_no_graphviz(msg, have_graphviz=False):
    """Raise an error when Graphviz cannot be found."""
    if not have_graphviz:
        raise ImportError('Graphviz python lib could not be imported: ' + msg)



"""An evaluation context for metrics.
"""
from __future__ import unicode_literals, print_function

import pandas as pd
from cyclus import lib

from cymetric.tools import raw_to_series


METRIC_REGISTRY = {}


def register_metric(cls):
    """Adds a metric to the registry."""
    METRIC_REGISTRY[cls.__name__] = cls


class Evaluator(object):
    """An evaluation context for metrics."""

    def __init__(self, db, write=True):
        """Parameters
        ----------
        db : database
        write : bool, optional
            Flag for whether metrics should be written to the database.

        Attributes
        ----------
        metrics : dict
            Metric instances bound the evaluator's database.
        rawcache : dict
            Results of querying metrics with given conditions.
        """
        self.write = write
        self.metrics = {}
        self.rawcache = {}
        self.db = db
        self.recorder = rec = lib.Recorder(inject_sim_id=False)
        rec.register_backend(db)
        self.known_tables = db.tables

    def get_metric(self, metric):
        """Checks if metric is already in the registry; adds it if not."""
        if metric not in self.metrics:
            self.metrics[metric] = METRIC_REGISTRY[metric](self.db)
        return self.metrics[metric]

    def eval(self, metric, conds=None):
        """Evalutes a metric with the given conditions."""
        rawkey = (metric, conds if conds is None else frozenset(conds))
        if rawkey in self.rawcache:
            return self.rawcache[rawkey]
        m = self.get_metric(metric)
        frames = []
        for dep in m.dependencies:
            frame = self.eval(dep, conds=conds)
            frames.append(frame)
        raw = m(frames=frames, conds=conds, known_tables=self.known_tables)
        if raw is None:
            return raw
        self.rawcache[rawkey] = raw
        # write back to db
        if (m.name in self.known_tables) or (not self.write):
            return raw
        rec = self.recorder
        rawd = raw.to_dict(orient='list')
        for i in range(len(raw)):
            d = rec.new_datum(m.name)
            for field, dbtype, shape in m.schema:
                fname = m.schema.byte_names[field]
                val = rawd[str(field)][i]
                d = d.add_val(fname, val, type=dbtype, shape=shape)
            d.record()
        self.known_tables.add(m.name)
        return raw


def eval(metric, db, conds=None, write=True):
    """Evalutes a metric with the given conditions in a database."""
    e = Evaluator(db, write=write)
    return e.eval(str(metric), conds=conds)



"""Root metrics for Bright-lite"""
from cymetric.root_metrics import root_metric


agentstate_brightlite_fuelfabfacilityinfo = root_metric(
    name='AgentState_Brightlite_FuelfabFacilityInfo')
agentstate_brightlite_reactorfacilityinfo = root_metric(
    name='AgentState_Brightlite_ReactorFacilityInfo')
agentstate_brightlite_reprocessfacilityinfo = root_metric(
    name='AgentState_Brightlite_ReprocessFacilityInfo')
brightlite_reactor_data = root_metric(
    name='BrightLite_Reactor_Data')



"""A plot generator for Cymetric.
"""
import warnings

import pandas as pd
import numpy as np

try:
    from graphviz import Digraph
    HAVE_GRAPHVIZ = True
except ImportError:
    HAVE_GRAPHVIZ = False

try:
    from pyne import data
    import pyne.enrichment as enr
    from pyne import nucname
    HAVE_PYNE = True
except ImportError:
    HAVE_PYNE = False

from cymetric import tools
from cymetric.filters import transactions_nuc


def flow_graph(evaler, senders=(), receivers=(), commodities=(), nucs=(), start=None, stop=None):
    """
    Generate the dot graph of the transation between facilitiese. Applying times
    nuclides selection when required.

    Parameters
    ----------
    evaler : evaler
    senders : list of the sending facility to consider
    receivers : list of the receiving facility to consider
    commodities : list of the commodity exchanged to consider
    nucs : list of nuclide to consider
    start : first timestep to consider, start included
    stop : last timestep to consider, stop included
    """
    tools.raise_no_graphviz('Unable to generate flow graph!', HAVE_GRAPHVIZ)

    df = transactions_nuc(
        evaler, senders, receivers, commodities, nucs)

    if start != None:
        df = df.loc[(df['Time'] >= start)]
    if stop != None:
        df = df.loc[(df['Time'] <= stop)]

    group_end = ['ReceiverPrototype', 'SenderPrototype', 'Commodity']
    group_start = group_end + ['Mass']
    df = df[group_start].groupby(group_end).sum()
    df.reset_index(inplace=True)

    agents_ = evaler.eval('AgentEntry')['Prototype'].tolist()

    dot = Digraph('G')

    for agent in agents_:
        dot.node(agent)

    for index, row in df.iterrows():
        lbl = str(row['Commodity']) + " " + str('{:.2e}'.format(row['Mass']))
        dot.edge(row['SenderPrototype'], row['ReceiverPrototype'],
                 label= lbl)

    return dot



"""Schemas for cymetric.
"""
from __future__ import print_function, unicode_literals
import sys
from collections import MutableSequence, Iterable
if sys.version_info[0] > 2:
    str_types = (str, bytes)
else:
    str_types = (str, unicode)

from cyclus import typesystem as ts
from cyclus.typesystem import NAMES, RANKS, IDS


def canon_dbtype(dbtype):
    """Returns the canonical form for a Cyclus database type."""
    if dbtype in NAMES:
        rtn = dbtype
    elif dbtype in IDS:
        rtn = IDS[dbtype]
    else:
        msg = 'the Cyclus type of {0!r} could not be determined.'
        raise TypeError(msg.format(dbtype))
    return rtn


def canon_shape(shape, dbtype):
    """Returns the canonical shape for a Cyclus database type."""
    if shape is None:
        rtn = None
    elif isinstance(shape, Iterable):
        if dbtype not in RANKS:
            dbtype = canon_dbtype(dbtype)
        s = []
        for i in shape:
            i = int(i)
            if i < -1:
                msg = '{0!r} of shape {1!r} must be >= -1.'
                raise ValueError(msg.format(i, shape))
            s.append(i)
        rank = RANKS[dbtype]
        if len(s) != rank:
            msg = ('the shape rank for the {0} type must be {1}, got {2} for a '
                   'shape of {3}.')
            raise ValueError(msg.format(NAMES[dbtype], rank, len(s), s))
        rtn = tuple(s)
    elif isinstance(shape, int):
        rtn = canon_shape((shape,), dbtype)
    else:
        msg = 'the shape {0!r} for a Cyclus type of {1!r} could not be determined.'
        raise TypeError(msg.format(shape, dbtype))
    return rtn


def canon_name(name):
    """Returns canonical form of a column name."""
    if not isinstance(name, str_types):
        raise ValueError('column name {0!r} must be a string'.format(name))
    return name


def canon_column(col):
    """Returns canonical form of a column."""
    if len(col) == 2:
        name, dbtype = col
        shape = None
    elif len(col) == 3:
        name, dbtype, shape = col
    else:
        msg = '{0!r} does not have length 2 or 3, cannot convert to schema.'
        raise ValueError(msg.format(col))
    name = canon_name(name)
    dbtype = canon_dbtype(dbtype)
    shape = canon_shape(shape, dbtype)
    return (name, dbtype, shape)


class SchemaProperty(MutableSequence):
    """Represents specific properties of a schema."""

    def __init__(self, obj, prop):
        self.obj = obj
        if prop == 'names':
            i = 0
            c = canon_name
        elif prop == 'dbtypes':
            i = 1
            c = canon_dbtype
        elif prop == 'shapes':
            i = 2
            c = canon_shape
        self.idx = i
        self.canonizer = c

    def _newcol(self, col, val):
        val = self.canonizer(val)
        col = list(col)
        col[self.idx] = val
        return tuple(col)

    def __get__(self, obj, objtype):
        return self

    def __set__(self, obj, val):
        if len(obj) != len(val):
            msg = 'The length of the schema ({0}) does not match the given length {1}.'
            raise ValueError(msg.format(len(obj), len(val)))
        newdescr = []
        for col, v in zip(obj, val):
            newdescr.append(self._newcol(col, v))
        obj.descr[:] = newdescr
        if self.idx == 0:
            obj.byte_names.clear()
            obj.byte_names.update({name: name.encode() for name, _, _ in newdescr})

    def __getitem__(self, i):
        return self.obj.descr[i][self.idx]

    def __setitem__(self, i, val):
        if self.idx == 0:
            del obj.byte_names[self.obj.descr[i][0]]
            obj.byte_names[val] = val.encode()
        self.obj.descr[i] = col = self._newcol(self.obj.descr[i], val)

    def __delitem__(self, i):
        raise AttributeError("May not delete from schema property.")

    def __len__(self):
        return len(self.obj)

    def insert(self, i, val):
        raise AttributeError("May not insert from schema property.")

    def __iter__(self):
        for col in self.obj.descr:
            yield col[self.idx]

    def __str__(self):
        s = '[' + ', '.join([repr(x) for x in self]) + ']'
        return s


class schema(MutableSequence):
    """A type represented compound schema composed of column names,
    associated types from the Cyclus type system, and optional shapes.
    """

    def __init__(self, x):
        """Parameters
        ----------
        x : iterable
            Sequence of 2- or 3-tuples that represent the schema.
        """
        self.descr = [canon_column(val) for val in x]
        self.names = SchemaProperty(self, 'names')
        self.dbtypes = SchemaProperty(self, 'dbtypes')
        self.shapes = SchemaProperty(self, 'shapes')
        self.byte_names = {name: name.encode() for name in self.names}

    def __getitem__(self, i):
        return self.descr[i]

    def __setitem__(self, i, val):
        self.descr[i] = canon_column(val)

    def __delitem__(self, i):
        del self.descr[i]

    def __len__(self):
        return len(self.descr)

    def insert(self, i, val):
        self.descr.insert(i, canon_column(val))

    def __iter__(self):
        for col in self.descr:
            yield col

    def __str__(self):
        s = '[' + ', '.join([repr(x) for x in self]) + ']'
        return s



import warnings
import pandas as pd
import numpy as np

try:
    from graphviz import Digraph
    HAVE_GRAPHVIZ = True
except ImportError:
    HAVE_GRAPHVIZ = False

try:
    from pyne import data
    import pyne.enrichment as enr
    from pyne import nucname
    HAVE_PYNE = True
except ImportError:
    HAVE_PYNE = False


from cymetric.tools import format_nucs, reduce, merge, add_missing_time_step


def transactions(evaler, senders=(), receivers=(), commodities=()):
    """
    Filter the Transaction Data Frame on specific sending facility and
    receving facility.

    Parameters
    ----------
    evaler : evaler
    senders :  of the sending facility
    receivers :  of the receiving facility
    commodities :  of the commodity exchanged
    """

    # initiate evaluation
    trans = evaler.eval('Transactions')
    agents = evaler.eval('AgentEntry')

    rec_agent = agents.rename(index=str, columns={'AgentId': 'ReceiverId'})
    if len(receivers) != 0:
        rec_agent = rec_agent[rec_agent['Prototype'].isin(receivers)]

    send_agent = agents.rename(index=str, columns={'AgentId': 'SenderId'})
    if len(senders) != 0:
        send_agent = send_agent[send_agent['Prototype'].isin(senders)]

    # check if sender and receiver exist
    if rec_agent.empty or send_agent.empty:
        return None

    # Clean Transation PDF
    rdc_table = []
    rdc_table.append(['ReceiverId', rec_agent['ReceiverId'].tolist()])
    rdc_table.append(['SenderId', send_agent['SenderId'].tolist()])
    if len(commodities) != 0:
        rdc_table.append(['Commodity', commodities])

    trans = reduce(trans, rdc_table)

    # Merge Sender to Transaction PDF
    base_col = ['SimId', 'SenderId']
    added_col = base_col + ['Prototype']
    trans = merge(trans, base_col, send_agent, added_col)
    trans = trans.rename(index=str, columns={
                         'Prototype': 'SenderPrototype'})

    # Merge Receiver to Transaction PDF
    base_col = ['SimId', 'ReceiverId']
    added_col = base_col + ['Prototype']
    trans = merge(trans, base_col, rec_agent, added_col)
    trans = trans.rename(index=str, columns={
                         'Prototype': 'ReceiverPrototype'})

    return trans


def transactions_nuc(evaler, senders=(), receivers=(), commodities=(), nucs=()):
    """
    Filter the Transaction Data Frame, which include nuclide composition, on
    specific sending facility and receving facility. Applying nuclides
    selection when required.

    Parameters
    ----------
    evaler : evaler
    senders :  of the sending facility
    receivers :  of the receiving facility
    commodities :  of the commodity exchanged
    nucs :  of nuclide to select.
    """

    compo = evaler.eval('Materials')

    df = transactions(evaler, senders, receivers, commodities)

    if len(nucs) != 0:
        nucs = format_nucs(nucs)
        compo = reduce(compo, [['NucId', nucs]])

    base_col = ['SimId', 'ResourceId']
    added_col = base_col + ['NucId', 'Mass']
    df = merge(df, base_col, compo, added_col)

    return df


def transactions_activity(evaler, senders=(), receivers=(), commodities=(), nucs=()):
    """
    Return the transation df, with the activities. Applying nuclides selection
    when required.

    Parameters
    ----------
    evaler : evaler
    senders :  of the sending facility
    receivers :  of the receiving facility
    commodities :  of the commodity exchanged
    nucs :  of nuclide to select.
    """

    df = transactions(evaler, senders, receivers, commodities)

    if len(nucs) != 0:
        nucs = format_nucs(nucs)

    compo = evaler.eval('Activity')
    compo = reduce(compo, [['NucId', nucs]])

    base_col = ['SimId', 'ResourceId']
    added_col = base_col + ['NucId', 'Activity']
    df = merge(df, base_col, compo, added_col)

    return df


def transactions_decayheat(evaler, senders=(), receivers=(), commodities=(), nucs=()):
    """
    Return the transation df, with the decayheat. Applying nuclides selection
    when required.

    Parameters
    ----------
    evaler : evaler
    senders :  of the sending facility
    receivers :  of the receiving facility
    commodities :  of the commodity exchanged
    nucs :  of nuclide to select.
    """

    df = transactions(evaler, senders, receivers, commodities)

    if len(nucs) != 0:
        nucs = format_nucs(nucs)

    compo = evaler.eval('DecayHeat')
    compo = reduce(compo, [['NucId', nucs]])

    base_col = ['SimId', 'ResourceId']
    added_col = base_col + ['NucId', 'DecayHeat']
    df = merge(df, base_col, compo, added_col)

    return df


def inventories(evaler, facilities=(), nucs=()):
    """
    Shape the reduced inventory Data Frame. Applying nuclides/facilities selection when required.

    Parameters
    ----------
    evaler : evaler
    facilities :  of the facility
    nucs :  of nuclide to select.
    """

    # Get inventory table
    df = evaler.eval('ExplicitInventory')
    agents = evaler.eval('AgentEntry')

    rdc_table = []  # because we want to get rid of the nuclide asap
    if len(nucs) != 0:
        nucs = format_nucs(nucs)
        rdc_table.append(['NucId', nucs])

    if len(facilities) != 0:
        agents = agents[agents['Prototype'].isin(facilities)]
        rdc_table.append(['AgentId', agents['AgentId'].tolist()])
    else:
        wng_msg = "no faciity provided"
        warnings.warn(wng_msg, UserWarning)
    df = reduce(df, rdc_table)

    base_col = ['SimId', 'AgentId']
    added_col = base_col + ['Prototype']
    df = merge(df, base_col, agents, added_col)

    return df


def inventories_activity(evaler, facilities=(), nucs=()):
    """
    Get a simple time series of the activity of the inventory in the selcted
    facilities. Applying nuclides selection when required.

    Parameters
    ----------
    evaler : evaler
    facilities :  of the facility
    nucs :  of nuclide to select.
    """

    if len(nucs) != 0:
        nucs = format_nucs(nucs)

    df = inventories(evaler, facilities, nucs)
    for i, row in df.iterrows():
        val = 1000 * data.N_A * row['Quantity'] * \
            data.decay_const(int(row['NucId']))
        df.set_value(i, 'Activity', val)

    return df


def inventories_decayheat(evaler, facilities=(), nucs=()):
    """
    Get a Inventory PDF including the decay heat of the inventory in the selected
    facilities. Applying nuclides selection when required.

    Parameters
    ----------
    evaler : evaler
    facilities :  of the facility
    nucs :  of nuclide to select.
    """

    if len(nucs) != 0:
        nucs = format_nucs(nucs)

    df = inventories_activity(evaler, facilities, nucs)
    for i, row in df.iterrows():
        val = data.MeV_per_MJ * \
            row['Activity'] * data.q_val(int(row['NucId']))
        df.set_value(i, 'DecayHeat', val)

    return df



"""A collection of basic metrics coming from the database that are
generated by Cyclus itself.
"""
from __future__ import print_function, unicode_literals

from cymetric.evaluator import register_metric

def _genrootclass(name):
    """Creates a new root metric class."""
    class Cls(object):
        dependencies = ()

        @property
        def schema(self):
            """Defines schema for root metric if provided."""
            if self._schema is not None:
                return self._schema
            # fill in schema code

        @property
        def name(self):
            """Assigns a name to root metric object."""
            return self.__class__.__name__

        def __init__(self, db):
            """Constructor for root metric object in database."""
            self._schema = None
            self.db = db

        def __call__(self, conds=None, *args, **kwargs):
            """Retrieves a root metric with given input conditions."""
            if self.name not in self.db.tables:
                return None
            return self.db.query(self.name, conds=conds)

    Cls.__name__ = str(name)
    register_metric(Cls)
    return Cls


def root_metric(obj=None, name=None, schema=None, *args, **kwargs):
    """Decorator that creates a root metric from a function or class."""
    if obj is not None:
        raise RuntimeError
    if name is None:
        raise RuntimeError
    return _genrootclass(name=name)


#core tables
resources = root_metric(name='Resources')
compositions = root_metric(name='Compositions')
recipes = root_metric(name='Recipes')
products = root_metric(name='Products')
res_creators = root_metric(name='ResCreators')
transactions = root_metric(name='Transactions')
info = root_metric(name='Info')
time_step_dur = root_metric(name='TimeStepDur')
finish = root_metric(name='Finish')
input_files = root_metric(name='InputFiles')
decom_schedule = root_metric(name='DecomSchedule')
build_schedule = root_metric(name='BuildSchedule')
snapshots = root_metric(name='Snapshots')
debug_requests = root_metric(name='DebugRequests')
debug_bids = root_metric(name='DebugBids')
explicit_inventory = root_metric(name='ExplicitInventory')
explicit_inventory_compact = root_metric(name='ExplicitInventoryCompact')

#where do these tables come from?
commod_priority = root_metric(name='CommodPriority')
decay_mode = root_metric(name='DecayMode')
field_types = root_metric(name='FieldTypes')
material_info = root_metric(name='MaterialInfo')
next_ids = root_metric(name='NextIds')
prototypes = root_metric(name='Prototypes')
xmlpp_info = root_metric(name='XMLPPInfo')

#general agent state tables
agent_entry = root_metric(name='AgentEntry')
agent_exit = root_metric(name='AgentExit')
agent_versions = root_metric(name='AgentVersions')
agentstate_agent = root_metric(name='AgentStateAgent')
agentstate_inventories = root_metric(name='AgentStateInventories')

#tables about solvers
greedy_solver_info = root_metric(name='GreedySolverInfo')
exchange_solver_info = root_metric(name='ExchangeSolverInfo')
prog_solver_info = root_metric(name='ProgSolverInfo')
solver_info = root_metric(name='SolverInfo')

#cyclus archetype tables
agentstate_agents_nullinstinfo = root_metric(
    name='AgentState_agents_NullInstInfo')
agentstate_agents_nullregioninfo = root_metric(
    name='AgentState_agents_NullRegionInfo')
agentstate_agent_sinkinfo = root_metric(
    name='AgentState_agents_SinkInfo')
agentstate_agent_sourceinfo = root_metric(
    name='AgentState_agents_SourceInfo')

#toolkit-enabled tables
time_series_power = root_metric(name='TimeSeriesPower')
time_series_enrichmentfeed = root_metric(name='TimeSeriesEnrichmentFeed')
time_series_enrichmentswu = root_metric(name='TimeSeriesEnrichmentSWU')



import warnings

import pandas as pd
import numpy as np

try:
    from pyne import data
    import pyne.enrichment as enr
    from pyne import nucname
    HAVE_PYNE = True
except ImportError:
    HAVE_PYNE = False


from cymetric import tools
from cymetric import filters


def transactions(evaler, senders=(), receivers=(), commodities=(), nucs=()):
    """
    Shape the reduced transation Data Frame into a simple time serie. Applying nuclides selection when required.

    Parameters
    ----------
    evaler : evaler
    senders :  of the sending facility
    receivers :  of the receiving facility
    commodities :  of the commodity exchanged
    nucs :  of nuclide to select.
    """

    if len(nucs) != 0:
        nucs = tools.format_nucs(nucs)

    df = filters.transactions_nuc(
        evaler, senders, receivers, commodities, nucs)

    group_end = ['ReceiverPrototype', 'SenderPrototype', 'Time']
    group_start = group_end + ['Mass']
    df = df[group_start].groupby(group_end).sum()
    df.reset_index(inplace=True)

    df = df[['Time', 'Mass']].groupby(['Time']).sum()
    df.reset_index(inplace=True)

    time = evaler.eval('TimeList')
    df = tools.add_missing_time_step(df, time)
    return df


def transactions_activity(evaler, senders=(), receivers=(), commodities=(), nucs=()):
    """
    Shape the reduced transation Data Frame into a simple time serie. Applying nuclides selection when required.

    Parameters
    ----------
    evaler : evaler
    senders :  of the sending facility
    receivers :  of the receiving facility
    commodities :  of the commodity exchanged
    nucs :  of nuclide to select.
    """

    if len(nucs) != 0:
        nucs = tools.format_nucs(nucs)

    df = filters.transactions_activity(evaler, senders, receivers, commodities,
                                     nucs)

    group_end = ['ReceiverPrototype', 'SenderPrototype', 'Time']
    group_start = group_end + ['Activity']
    df = df[group_start].groupby(group_end).sum()
    df.reset_index(inplace=True)

    df = df[['Time', 'Activity']].groupby(['Time']).sum()
    df.reset_index(inplace=True)

    time = evaler.eval('TimeList')
    df = tools.add_missing_time_step(df, time)
    return df


def transactions_decayheat(evaler, senders=(), receivers=(), commodities=(), nucs=()):
    """
    Shape the reduced transation Data Frame into a simple time serie. Applying nuclides selection when required.

    Parameters
    ----------
    evaler : evaler
    senders :  of the sending facility
    receivers :  of the receiving facility
    commodities :  of the commodity exchanged
    nucs :  of nuclide to select.
    """

    if len(nucs) != 0:
        nucs = tools.format_nucs(nucs)

    df = filters.transactions_decayheat(evaler, senders, receivers, commodities,
                                      nucs)

    group_end = ['ReceiverPrototype', 'SenderPrototype', 'Time']
    group_start = group_end + ['DecayHeat']
    df = df[group_start].groupby(group_end).sum()
    df.reset_index(inplace=True)

    df = df[['Time', 'DecayHeat']].groupby(['Time']).sum()
    df.reset_index(inplace=True)

    time = evaler.eval('TimeList')
    df = tools.add_missing_time_step(df, time)
    return df




def inventories(evaler, facilities=(), nucs=()):
    """
    Shape the reduced inventory Data Frame into a simple time serie. Applying
    nuclides/facilities selection when required.

    Parameters
    ----------
    evaler : evaler
    facilities :  of the facility
    nucs :  of nuclide to select.
    """

    if len(nucs) != 0:
        nucs = tools.format_nucs(nucs)
    else:
        wng_msg = "no nuclide provided"
        warnings.warn(wng_msg, UserWarning)

    df = filters.inventories(evaler, facilities, nucs)

    group_end = ['Time']
    group_start = group_end + ['Quantity']
    df = df[group_start].groupby(group_end).sum()
    df.reset_index(inplace=True)

    time = evaler.eval('TimeList')
    df = tools.add_missing_time_step(df, time)
    return df


def inventories_activity(evaler, facilities=(), nucs=()):
    """
    Get a simple time series of the decay heat of the inventory in the selcted
    facilities. Applying nuclides selection when required.

    Parameters
    ----------
    evaler : evaler
    facilities :  of the facility
    nucs :  of nuclide to select.
    """

    if len(nucs) != 0:
        nucs = tools.format_nucs(nucs)
    else:
        wng_msg = "no nuclide provided"
        warnings.warn(wng_msg, UserWarning)

    df = filters.inventories_activity(evaler, facilities, nucs)
    group_end = ['Time']
    group_start = group_end + ['Activity']
    df = df[group_start].groupby(group_end).sum()
    df.reset_index(inplace=True)

    time = evaler.eval('TimeList')
    df = tools.add_missing_time_step(df, time)
    return df


def inventories_decayheat(evaler, facilities=(), nucs=()):
    """
    Get a simple time series of the decay heat of the inventory in the selcted
    facilities. Applying nuclides selection when required.

    Parameters
    ----------
    evaler : evaler
    facilities :  of the facility
    nucs :  of nuclide to select.
    """

    if len(nucs) != 0:
        nucs = tools.format_nucs(nucs)
    else:
        wng_msg = "no nuclide provided"
        warnings.warn(wng_msg, UserWarning)

    df = filters.inventories_decayheat(evaler, facilities, nucs)
    group_end = ['Time']
    group_start = group_end + ['DecayHeat']
    df = df[group_start].groupby(group_end).sum()
    df.reset_index(inplace=True)

    time = evaler.eval('TimeList')
    df = tools.add_missing_time_step(df, time)
    return df


def get_power(evaler, facilities=()):
    """
    Shape the reduced Power Data Frame into a simple time serie. Applying
    facilities selection when required.

    Parameters
    ----------
    evaler : evaler
    facilities :  of the facility
    """

    # Get inventory table
    power = evaler.eval('TimeSeriesPower')
    agents = evaler.eval('AgentEntry')

    rdc_ = []  # because we want to get rid of the facility asap
    if len(facilities) != 0:
        agents = agents[agents['Prototype'].isin(facilities)]
        rdc_.append(['AgentId', agents['AgentId'].tolist()])
    else:
        wng_msg = "no faciity provided"
        warnings.warn(wng_msg, UserWarning)
    power = tools.reduce(power, rdc_)

    base_col = ['SimId', 'AgentId']
    added_col = base_col + ['Prototype']
    power = tools.merge(power, base_col, agents, added_col)

    group_end = ['Time']
    group_start = group_end + ['Value']
    df = power[group_start].groupby(group_end).sum()
    df.reset_index(inplace=True)

    time = evaler.eval('TimeList')
    df = tools.add_missing_time_step(df, time)
    return df


def get_deployment(evaler, facilities=()):
    """
    Get a simple time series with deployment schedule of the selected facilities.

    Parameters
    ----------
    evaler : evaler
    facilities :  of the facility
    """

    # Get inventory table
    df = evaler.eval('AgentEntry')

    rdc_ = []  # because we want to get rid of the facility asap
    if len(facilities) != 0:
        df = df[df['Prototype'].isin(facilities)]
        rdc_.append(['AgentId', df['AgentId'].tolist()])
    else:
        wng_msg = "no faciity provided"
        warnings.warn(wng_msg, UserWarning)

    # Adding a constant column to easily sum the amount of facilities build per
    # time step
    df = df.assign(Value=lambda x: 1)
    group_end = ['EnterTime']
    group_start = group_end + ['Value']
    df = df[group_start].groupby(group_end).sum()
    df.reset_index(inplace=True)
    df.rename(index=str, columns={'EnterTime': 'Time'}, inplace=True)

    time = evaler.eval('TimeList')
    df = tools.add_missing_time_step(df, time)
    return df


def get_retirement(evaler, facilities=()):
    """
    Get a simple time series with retirement schedule of the selected facilities.

    Parameters
    ----------
    evaler : evaler
    facilities :  of the facility
    """

    # Get inventory table
    df = evaler.eval('AgentEntry')
    df = df[df['Lifetime'] > 0]

    rdc_ = []  # because we want to get rid of the facility asap
    if len(facilities) != 0:
        df = df[df['Prototype'].isin(facilities)]
        rdc_.append(['AgentId', df['AgentId'].tolist()])
    else:
        wng_msg = "no faciity provided"
        warnings.warn(wng_msg, UserWarning)

    # Adding a constant column to easily sum the amount of facilities build per
    # time step i

    df = df.assign(Value=lambda x: 1)
    df['DecomTime'] = df['EnterTime'] + df['Lifetime']
    group_end = ['DecomTime']
    group_start = group_end + ['Value']
    df = df[group_start].groupby(group_end).sum()
    df.reset_index(inplace=True)

    df.rename(index=str, columns={'DecomTime': 'Time'}, inplace=True)
    time = evaler.eval('TimeList')
    df = tools.add_missing_time_step(df, time)
    return df



"""Cymetric: The Cyclus Analysis Toolkit"""
from __future__ import unicode_literals, print_function

from cyclus.typesystem import *  # only grabs code generated defintiions
from cyclus.lib import Datum, FullBackend, SqliteBack, Hdf5Back, \
        Recorder

try:
    from cymetric.tools import dbopen
    from cymetric.schemas import schema, canon_dbtype, canon_shape, \
        canon_column, canon_name
    from cymetric.root_metrics import root_metric
    from cymetric.metrics import Metric, metric
    from cymetric.evaluator import METRIC_REGISTRY, register_metric, \
        raw_to_series, Evaluator, eval
    from cymetric.execution import ExecutionContext, exec_code
except ImportError:
    # again with the wacky CI issues
    from .tools import dbopen
    from .schemas import schema, canon_dbtype, canon_shape, \
        canon_column, canon_name
    from .root_metrics import root_metric
    from .metrics import Metric, metric
    from .evaluator import METRIC_REGISTRY, register_metric, \
        raw_to_series, Evaluator, eval
    from .execution import ExecutionContext, exec_code

__version__ = '1.5.3'




"""The CLI for cymetric."""
from __future__ import unicode_literals, print_function
import os
import sys
from argparse import ArgumentParser, Namespace

from cymetric.tools import dbopen
from cymetric.evaluator import Evaluator
from cymetric.execution import exec_code

def parse_args():
    """Parses the command line arguments."""
    parser =  ArgumentParser(description='Cyclus metric analysis tool.')
    parser.add_argument('db', help='path to the database')
    parser.add_argument('-e', dest='exec_code', help='execution string',
                        default=None)
    parser.add_argument('-l', dest='listing', action='store_true', 
                        help='lists the tables in the database', default=False)
    parser.add_argument('--write', dest='write', action='store_true',
                        help='writes the computed metrics to the db',
                        default=True)
    parser.add_argument('--no-write', dest='write', action='store_false',
                        help='does not write the computed metrics to the db')
    ns = parser.parse_args()
    return ns


def main():
    """CLI entry point."""
    ns = parse_args()
    db = dbopen(ns.db)
    if ns.listing:
        for tab in sorted(db.tables):
            print(tab)
    if ns.exec_code is not None:
        exec_code(ns.exec_code, db, write=ns.write)

if __name__ == '__main__': 
    main()



"""A collection of metrics that come stock with cymetric.
"""
from __future__ import print_function, unicode_literals
import inspect

import numpy as np
import pandas as pd

try:
    from pyne import data
    import pyne.enrichment as enr
    HAVE_PYNE = True
except ImportError:
    HAVE_PYNE = False

from cyclus import lib
from cyclus import typesystem as ts

try:
    from cymetric import schemas
    from cymetric import tools
    from cymetric.evaluator import register_metric
except ImportError:
    # some wacky CI paths prevent absolute importing, try relative
    from . import schemas
    from . import tools
    from .evaluator import register_metric


class Metric(object):
    """Metric class"""
    dependencies = NotImplemented
    schema = NotImplemented

    def __init__(self, db):
        self.db = db

    @property
    def name(self):
        return self.__class__.__name__


def _genmetricclass(f, name, depends, scheme):
    """Creates a new metric class with a given name, dependencies, and schema.

    Parameters
    ----------
    name : str
        Metric name
    depends : list of lists (table name, tuple of indices, column name)
        Dependencies on other database tables (metrics or root metrics)
    scheme : list of tuples (column name, data type)
        Schema for metric
    """
    if not isinstance(scheme, schemas.schema):
        scheme = schemas.schema(scheme)

    class Cls(Metric):
        dependencies = depends
        schema = scheme
        func = staticmethod(f)

        __doc__ = inspect.getdoc(f)

        def __init__(self, db):
            """Constructor for metric object in database."""
            super(Cls, self).__init__(db)

        def __call__(self, frames, conds=None, known_tables=None, *args, **kwargs):
            """Computes metric for given input data and conditions."""
            # FIXME test if I already exist in the db, read in if I do
            if known_tables is None:
                known_tables = self.db.tables()
            if self.name in known_tables:
                return self.db.query(self.name, conds=conds)
            return f(*frames)

    Cls.__name__ = str(name)
    register_metric(Cls)
    return Cls


def metric(name=None, depends=NotImplemented, schema=NotImplemented):
    """Decorator that creates metric class from a function or class."""
    def dec(f):
        clsname = name or f.__name__
        return _genmetricclass(f=f, name=clsname, scheme=schema, depends=depends)
    return dec


#####################
## General Metrics ##
#####################

# Material Mass (quantity * massfrac)
_matdeps = ['Resources', 'Compositions']

_matschema = [
    ('SimId', ts.UUID),
    ('QualId', ts.INT),
    ('ResourceId', ts.INT),
    ('ObjId', ts.INT),
    ('TimeCreated', ts.INT),
    ('NucId', ts.INT),
    ('Units', ts.STRING),
    ('Mass', ts.DOUBLE)
    ]

@metric(name='Materials', depends=_matdeps, schema=_matschema)
def materials(rsrcs, comps):
    """Materials metric returns the material mass (quantity of material in
    Resources times the massfrac in Compositions) indexed by the SimId, QualId,
    ResourceId, ObjId, TimeCreated, and NucId.
    """
    x = pd.merge(rsrcs, comps, on=['SimId', 'QualId'], how='inner')
    x = x.set_index(['SimId', 'QualId', 'ResourceId', 'ObjId','TimeCreated',
                     'NucId', 'Units'])
    y = x['Quantity'] * x['MassFrac']
    y.name = 'Mass'
    z = y.reset_index()
    return z


del _matdeps, _matschema


# Activity (mass * decay_const / atomic_mass)
_actdeps = ['Materials']

_actschema = [
    ('SimId', ts.UUID),
    ('QualId', ts.INT),
    ('ResourceId', ts.INT),
    ('ObjId', ts.INT),
    ('TimeCreated', ts.INT),
    ('NucId', ts.INT),
    ('Activity', ts.DOUBLE)
    ]

@metric(name='Activity', depends=_actdeps, schema=_actschema)
def activity(mats):
    """Activity metric returns the instantaneous activity of a nuclide
    in a material (material mass * decay constant / atomic mass)
    indexed by the SimId, QualId, ResourceId, ObjId, TimeCreated, and NucId.
    """
    tools.raise_no_pyne('Activity could not be computed', HAVE_PYNE)
    mass = tools.raw_to_series(mats,
                               ('SimId', 'QualId', 'ResourceId', 'ObjId', 'TimeCreated', 'NucId'),
                               'Mass')
    act = []
    for (simid, qual, res, obj, time, nuc), m in mass.iteritems():
        val = (1000 * data.N_A * m * data.decay_const(nuc) \
              / data.atomic_mass(nuc))
        act.append(val)
    act = pd.Series(act, index=mass.index)
    act.name = 'Activity'
    rtn = act.reset_index()
    return rtn

del _actdeps, _actschema


# DecayHeat (activity * q_value)
_dhdeps = ['Activity']

_dhschema = [
    ('SimId', ts.UUID),
    ('QualId', ts.INT),
    ('ResourceId', ts.INT),
    ('ObjId', ts.INT),
    ('TimeCreated', ts.INT),
    ('NucId', ts.INT),
    ('DecayHeat', ts.DOUBLE)
    ]

@metric(name='DecayHeat', depends=_dhdeps, schema=_dhschema)
def decay_heat(acts):
    """Decay heat metric returns the instantaneous decay heat of a nuclide
    in a material (Q value * activity) indexed by the SimId, QualId,
    ResourceId, ObjId, TimeCreated, and NucId.
    """
    tools.raise_no_pyne('DecayHeat could not be computed', HAVE_PYNE)
    act = tools.raw_to_series(acts,
                              ('SimId', 'QualId', 'ResourceId', 'ObjId', 'TimeCreated', 'NucId'),
                              'Activity')
    dh = []
    for (simid, qual, res, obj, time, nuc), a in act.iteritems():
        val = (data.MeV_per_MJ * a * data.q_val(nuc))
        dh.append(val)
    dh = pd.Series(dh, index=act.index)
    dh.name = 'DecayHeat'
    rtn = dh.reset_index()
    return rtn

del _dhdeps, _dhschema


# Agent Building
_bsdeps = ['AgentEntry']

_bsschema = [
    ('SimId', ts.UUID), ('EnterTime', ts.INT), ('Prototype', ts.STRING),
    ('Count', ts.INT)
    ]

@metric(name='BuildSeries', depends=_bsdeps, schema=_bsschema)
def build_series(entry):
    """Provides a time series of the building of agents by prototype.
    """
    entry_index = ['SimId', 'EnterTime', 'Prototype']
    count = entry.groupby(entry_index).size()
    count.name = 'Count'
    rtn = count.reset_index()
    return rtn


del _bsdeps, _bsschema


# Agent Decommissioning
_dsdeps = ['AgentEntry', 'AgentExit']

_dsschema = [
    ('SimId', ts.UUID),
    ('ExitTime', ts.INT),
    ('Prototype', ts.STRING),
    ('Count', ts.INT)
    ]

@metric(name='DecommissionSeries', depends=_dsdeps, schema=_dsschema)
def decommission_series(entry, exit):
    """Provides a time series of the decommissioning of agents by prototype.
    """
    exit_index = ['SimId', 'ExitTime', 'Prototype']
    if exit is not None:
        exit = pd.merge(entry, exit, on=['SimId', 'AgentId'], how='inner')
        exit = exit.set_index(exit_index)
    else:
        return print('No agents were decommissioned during this simulation.')
    count = exit.groupby(level=exit_index).size()
    count.name = 'Count'
    rtn = count.reset_index()
    return rtn

del _dsdeps, _dsschema


# Agents
_agentsdeps = ['AgentEntry', 'AgentExit', 'DecomSchedule', 'Info']

_agentsschema = schemas.schema([
    ('SimId', ts.UUID),
    ('AgentId', ts.INT),
    ('Kind', ts.STRING),
    ('Spec', ts.STRING),
    ('Prototype', ts.STRING),
    ('ParentId', ts.INT),
    ('Lifetime', ts.INT),
    ('EnterTime', ts.INT),
    ('ExitTime', ts.INT),
    ])

@metric(name='Agents', depends=_agentsdeps, schema=_agentsschema)
def agents(entry, exit, decom, info):
    """Computes the Agents table. This is tricky because both the AgentExit
    table and the DecomSchedule table may not be present in the database.
    Furthermore, the Info table does not contain the AgentId column. This
    computation handles the calculation of the ExitTime in the face a
    significant amounts of missing data.
    """
    mergeon = ['SimId', 'AgentId']
    ent = tools.raw_to_series(entry, ['SimId', 'AgentId'], 'Kind')
    idx = ent.index
    df = entry[['SimId', 'AgentId', 'Kind', 'Spec', 'Prototype', 'ParentId',
                'Lifetime', 'EnterTime']]
    if exit is None:
        agent_exit = pd.Series(index=idx, data=[np.nan]*len(idx))
        agent_exit.name = 'ExitTime'
    else:
        agent_exit = agent_exit.reindex(index=idx)
    df = pd.merge(df, agent_exit.reset_index(), on=mergeon)
    if decom is not None:
        df = tools.merge_and_fillna_col(df, decom[['SimId', 'AgentId', 'DecomTime']],
                                        'ExitTime', 'DecomTime', on=mergeon)
    df = tools.merge_and_fillna_col(df, info[['SimId', 'Duration']],
                                    'ExitTime', 'Duration', on=['SimId'])
    return df

del _agentsdeps, _agentsschema


# Transaction Quantity
_transdeps = ['Materials', 'Transactions']

_transschema = [
    ('SimId', ts.UUID),
    ('TransactionId', ts.INT),
    ('ResourceId', ts.INT),
    ('ObjId', ts.INT),
    ('TimeCreated', ts.INT),
    ('SenderId', ts.INT),
    ('ReceiverId', ts.INT),
    ('Commodity', ts.STRING),
    ('Units', ts.STRING),
    ('Quantity', ts.DOUBLE)
    ]

@metric(name='TransactionQuantity', depends=_transdeps, schema=_transschema)
def transaction_quantity(mats, tranacts):
    """Transaction Quantity metric returns the quantity of each transaction throughout
    the simulation.
    """
    trans_index = ['SimId', 'TransactionId', 'ResourceId', 'ObjId',
            'TimeCreated', 'SenderId', 'ReceiverId', 'Commodity', 'Units']
    trans = pd.merge(mats, tranacts, on=['SimId', 'ResourceId'], how='inner')
    trans = trans.set_index(trans_index)
    trans = trans.groupby(level=trans_index)['Mass'].sum()
    trans.name = 'Quantity'
    rtn = trans.reset_index()
    return rtn

del _transdeps, _transschema


# Explicit Inventory By Agent
_invdeps = ['ExplicitInventory']

_invschema = [
    ('SimId', ts.UUID),
    ('AgentId', ts.INT),
    ('Time', ts.INT),
    ('InventoryName', ts.STRING),
    ('NucId', ts.INT),
    ('Quantity', ts.DOUBLE)
    ]

@metric(name='ExplicitInventoryByAgent', depends=_invdeps, schema=_invschema)
def explicit_inventory_by_agent(expinv):
    """The Inventory By Agent metric groups the inventories by Agent
    (keeping all nuc information)
    """
    inv_index = ['SimId', 'AgentId', 'Time', 'InventoryName', 'NucId']
    inv = tools.raw_to_series(expinv,
                              ['SimId', 'AgentId', 'Time', 'InventoryName', 'NucId'],
                              'Quantity')
    inv = inv.groupby(level=inv_index).sum()
    inv.name = 'Quantity'
    rtn = inv.reset_index()
    return rtn

del _invdeps, _invschema


# Explicit Inventory By Nuc
_invdeps = ['ExplicitInventory']

_invschema = [
    ('SimId', ts.UUID),
    ('Time', ts.INT),
    ('InventoryName', ts.STRING),
    ('NucId', ts.INT),
    ('Quantity', ts.DOUBLE)
    ]

@metric(name='ExplicitInventoryByNuc', depends=_invdeps, schema=_invschema)
def explicit_inventory_by_nuc(expinv):
    """The Inventory By Nuc metric groups the inventories by nuclide
    and discards the agent information it is attached to (providing fuel
    cycle-wide nuclide inventories)
    """
    inv_index = ['SimId', 'Time', 'InventoryName', 'NucId']
    inv = tools.raw_to_series(expinv,
                              ('SimId', 'Time', 'InventoryName', 'NucId'),
                              'Quantity')
    inv = inv.groupby(level=inv_index).sum()
    inv.name = 'Quantity'
    rtn = inv.reset_index()
    return rtn

del _invdeps, _invschema


# Electricity Generated [MWe-y]
_egdeps = ['TimeSeriesPower']

_egschema = [
    ('SimId', ts.UUID),
    ('AgentId', ts.INT),
    ('Year', ts.INT),
    ('Energy', ts.DOUBLE)
    ]

@metric(name='AnnualElectricityGeneratedByAgent', depends=_egdeps, schema=_egschema)
def annual_electricity_generated_by_agent(elec):
    """Annual Electricity Generated metric returns the total electricity
    generated in MWe-y for each agent, calculated from the average monthly
    power given in TimeSeriesPower.
    """
    elec = pd.DataFrame(data={'SimId': elec.SimId,
                              'AgentId': elec.AgentId,
                              'Year': elec.Time.apply(lambda x: x//12),
                              'Energy': elec.Value.apply(lambda x: x/12)},
			columns=['SimId', 'AgentId', 'Year', 'Energy'])
    el_index = ['SimId', 'AgentId', 'Year']
    elec = elec.groupby(el_index).sum()
    rtn = elec.reset_index()
    return rtn

del _egdeps, _egschema

#
# Not a metric, not a root metric metrics
#

# These are not metrics that have any end use in mind, but are required for the
# calculation of other metrics. Required tables like this should be stored
# elsewhere in the future if they become more common.

# TimeList

_tldeps = ['Info']

_tlschema = [
    ('SimId', ts.UUID),
    ('TimeStep', ts.INT)
    ]

@metric(name='TimeList', depends=_tldeps, schema=_tlschema)
def timelist(info):
    """In case the sim does not have entries for every timestep, this populates
    a list with all timesteps in the duration.
    """
    info = tools.raw_to_series(info, ('SimId',), 'Duration')
    tl = []
    for sim, dur in info.iteritems():
        for i in range(dur):
            tl.append((sim, i))
    tl = pd.DataFrame(tl, columns=['SimId', 'TimeStep'])
    return tl


del _tldeps, _tlschema




"""FCO metrics"""
import pandas as pd
try:
    from pyne import data
    import pyne.enrichment as enr
    HAVE_PYNE = True
except ImportError:
    HAVE_PYNE = False
from cyclus import typesystem as ts

from cymetric import metric
from cymetric import tools

#########################
## FCO-related metrics ##
#########################


# Electricity Generated [GWe-y]
_egdeps = ['AnnualElectricityGeneratedByAgent']

_egschema = [('Year', ts.INT), ('Energy', ts.DOUBLE)]

@metric(name='FcoElectricityGenerated', depends=_egdeps, schema=_egschema)
def fco_electricity_generated(elec):
    """FcoElectricityGenerated metric returns the electricity generated in GWe-y
    for all agents in simulation.
    """
    elec = pd.DataFrame(data={'Year': elec.Year,
                              'Energy': elec.Energy.apply(lambda x: x/1000)},
                        columns=['Year', 'Energy'])
    elec = elec.groupby('Year').sum()
    rtn = elec.reset_index()
    return rtn

del _egdeps, _egschema


# U Resources Mined [t]
_udeps= ['Materials', 'Transactions']

_uschema = [('Year', ts.INT), ('UMined', ts.DOUBLE)]

@metric(name='FcoUMined', depends=_udeps, schema=_uschema)
def fco_u_mined(mats, trans):
    """FcoUMined metric returns the uranium mined in tonnes for each year
    in a simulation. This is written for simulations that use the
    Bright-lite Fuel Fab (i.e., the U235 and U238 are given separately in the
    FCO simulations, and the filtering is archetype-specific).
    """
    tools.raise_no_pyne('U_Mined could not be computed', HAVE_PYNE)
    mass = pd.merge(mats, trans, on=['ResourceId'], how='inner')
    mass = mass.set_index(['ObjId', 'TimeCreated', 'NucId'])
    u = []
    prods = {}
    mass235 = {}
    m = mass[mass['Commodity'] == 'LWR Fuel']
    for (obj, _, nuc), value in m.iterrows():
        if 922320000 <= nuc <= 922390000:
            prods[obj] = prods.get(obj, 0.0) + value['Mass']
        if nuc==922350000:
            mass235[obj] = value['Mass']
    x_feed = 0.0072
    x_tails = 0.0025
    for obj, m235 in mass235.items():
        x_prod = m235 / prods[obj]
        feed = enr.feed(x_feed, x_prod, x_tails, product=prods[obj]) / 1000
        u.append(feed)
    m = m.groupby(level=['ObjId', 'TimeCreated'])['Mass'].sum()
    m = m.reset_index()
    # sum by years (12 time steps)
    u = pd.DataFrame(data={'Year': m.TimeCreated.apply(lambda x: x//12),
                           'UMined': u}, columns=['Year', 'UMined'])
    u = u.groupby('Year').sum()
    rtn = u.reset_index()
    return rtn

del _udeps, _uschema


# SWU Required [million SWU]
_swudeps = ['Materials', 'Transactions']

_swuschema = [('Year', ts.INT), ('SWU', ts.DOUBLE)]

@metric(name='FcoSwu', depends=_swudeps, schema=_swuschema)
def fco_swu(mats, trans):
    """FcoSwu metric returns the separative work units required for each
    year in a simulation. This is written for simulations that
    use the Bright-lite Fuel Fab (i.e., the U235 and U238 are given separately
    in the FCO simulations, and the filtering is archetype-specific).
    """
    tools.raise_no_pyne('SWU Required could not be computed', HAVE_PYNE)
    mass = pd.merge(mats, trans, on=['ResourceId'], how='inner')
    mass = mass.set_index(['ObjId', 'TimeCreated', 'NucId'])
    swu = []
    prods = {}
    mass235 = {}
    m = mass[mass['Commodity'] == 'LWR Fuel']
    for (obj, _, nuc), value in m.iterrows():
        if 922320000 <= nuc <= 922390000:
            prods[obj] = prods.get(obj, 0.0) + value['Mass']
        if nuc == 922350000:
            mass235[obj] = value['Mass']
    x_feed = 0.0072
    x_tails = 0.0025
    for obj, m235 in mass235.items():
        x_prod = m235 / prods[obj]
        swu0 = enr.swu(x_feed, x_prod, x_tails, product=prods[obj]) / 1e6
        swu.append(swu0)
    m = m.groupby(level=['ObjId', 'TimeCreated'])['Mass'].sum()
    m = m.reset_index()
    # sum by years (12 time steps)
    swu = pd.DataFrame(data={'Year': m.TimeCreated.apply(lambda x: x//12),
                             'SWU': swu}, columns=['Year', 'SWU'])
    swu = swu.groupby('Year').sum()
    rtn = swu.reset_index()
    return rtn

del _swudeps, _swuschema


# Annual Fuel Loading Rate [tHM/y]
_fldeps = ['Materials', 'Transactions']

_flschema = [('Year', ts.INT), ('FuelLoading', ts.DOUBLE)]

@metric(name='FcoFuelLoading', depends=_fldeps, schema=_flschema)
def fco_fuel_loading(mats, trans):
    """FcoFuelLoading metric returns the fuel loaded in tHM/y in a
    simulation. This is written for FCO databases that use Bright-lite
    archetypes (the commodity filtering is specific to these archetypes).
    """
    mass = pd.merge(mats, trans, on=['ResourceId'], how='inner')
    mass = mass.set_index(['TimeCreated'])
    mass = mass.query('Commodity == ["LWR Fuel", "FR Fuel"]')
    mass = mass.groupby(mass.index)['Mass'].sum()
    mass = mass.reset_index()
    # sum by years (12 time steps)
    mass = pd.DataFrame(data={'Year': mass.TimeCreated.apply(lambda x: x//12),
                              'FuelLoading': mass.Mass.apply(lambda x: x/1000)},
                        columns=['Year', 'FuelLoading'])
    mass = mass.groupby('Year').sum()
    rtn = mass.reset_index()
    return rtn

del _fldeps, _flschema





#!/usr/bin/env python
"""Welcome to cymetric's setup.py script. This is a little non-standard because pyne
is a multilanguage project.  Still, this script follows a predicatable ordering:

1. Parse command line arguments
2. Call cmake from the 'build' directory
3. Call make from the 'build' directory
4. Use distuitls/setuptools from the 'build' directory

This gives us the best of both worlds. Compiled code is installed with cmake/make
and Cython/Python code is installed with normal Python tools. The only trick here is
how the various command line arguments are handed off to the three sub-processes.

To accomplish this we use argparser groups to group command line arguments based on
whether they go to:

1. the setup() function,
2. cmake,
3. make, or
4. other - typically used for args that apply to multiple other groups or
   modify the environment in some way.

To add a new command line argument, first add it to the appropriate group in the
``parse_args()`` function.  Then, modify the logic in the cooresponding
``parse_setup()``, ``parse_cmake()``, ``parse_make()``, or ``parse_others()``
functions to consume your new command line argument.  It is OK for more than
one of the parser functions to consume the argument. Where appropriate,
ensure the that argument is appended to the argument list that is returned by these
functions.
"""
from __future__ import print_function

import os
import sys
import imp
import argparse
import platform
import warnings
import subprocess
from glob import glob
from distutils import core, dir_util


VERSION = '1.5.3'
IS_NT = os.name == 'nt'


def main():
    scripts = [os.path.join('scripts', f) for f in os.listdir('scripts')]
    scripts = [s for s in scripts if ((IS_NT and s.endswith('.bat'))
                                      or (not IS_NT and
                                          not s.endswith('.bat')))]
    packages = ['cymetric']
    pack_dir = {'cymetric': 'cymetric'}
    setup_kwargs = {
        "name": "cymetric",
        "version": VERSION,
        "description": 'Cyclus Metric Calculator',
        "author": 'Cyclus Development Team',
        "author_email": 'cyclus-dev@googlegroups.com',
        "url": 'http://fuelcycle.org/',
        "packages": packages,
        "package_dir": pack_dir,
        "scripts": scripts,
        }
    rtn = core.setup(**setup_kwargs)
    return rtn


if __name__ == "__main__":
    main()



"""Root metrics for Cycamore"""
from cymetric import root_metric


#root metrics that come with the archetypes within the cycamore module
agentstate_cycamore_enrichmentinfo = root_metric(
    name='AgentState_cycamore_EnrichmentInfo')
agentstate_cycamore_fuelfabinfo = root_metric(
    name='AgentState_cycamore_FuelFabInfo')
agentstate_cycamore_reactorinfo = root_metric(
    name='AgentState_cycamore_ReactorInfo')
agentstate_cycamore_separationsinfo = root_metric(
    name='AgentState_cycamore_SeparationsInfo')
agentstate_cycamore_sinkinfo = root_metric(
    name='AgentState_cycamore_SinkInfo')
agentstate_cycamore_sourceinfo = root_metric(
    name='AgentState_cycamore_SourceInfo')
enrichments = root_metric(name='Enrichments')
reactor_events = root_metric(name='ReactorEvents')




"""Execution for cymetric.
"""
from __future__ import unicode_literals, print_function
import re
import sys
import uuid
from collections import MutableMapping, Sized
if sys.version_info[0] > 2:
    str_types = (str, bytes)
else:
    str_types = (str, unicode)

import numpy as np
import pandas as pd

from cymetric.evaluator import Evaluator, METRIC_REGISTRY


class ColumnProxy(object):
    """A proxy object for column that returns condition 3-tuples from
    comparison operations.
    """

    def __init__(self, name):
        """Parameters
        ----------
        name : str
            The column name.
        """
        self.name = name
        self.conds = []

    def __call__(self, *args, **kwargs):
        """Stub function that prevents ColumnProxies from being called."""
        raise TypeError('ColumnProxy object {0!r} is not callable'.format(self.name))

    def __getitem__(self, *args, **kwargs):
        """Stub function that prevents ColumnProxies from being indexed."""
        raise TypeError('ColumnProxy object {0!r} cannot be indexed'.format(self.name))

    def __lt__(self, other):
        """Less than operator for column filtering."""
        self.conds.append((self.name, '<', other))
        return self

    def __gt__(self, other):
        """Greater than operator for column filtering."""
        self.conds.append((self.name, '>', other))
        return self

    def __le__(self, other):
        """Less than or equal operator for column filtering."""
        self.conds.append((self.name, '<=', other))
        return self

    def __ge__(self, other):
        """Greater than or equal operator for column filtering."""
        self.conds.append((self.name, '>=', other))
        return self
        return self.name, '>=', other

    def __eq__(self, other):
        """Equal operator for column filtering."""
        self.conds.append((self.name, '==', other))
        return self

    def __ne__(self, other):
        """Not equal operator for column filtering."""
        self.conds.append((self.name, '!=', other))
        return self


stripper = lambda s: s.strip()

COND_RE = re.compile('\s*(\w+)\s*(<|>|<=|>=|==|!=)\s*(.*)')

def parse_cond(cond):
    """Parses a condition and returns the canonical 3-tuple."""
    if isinstance(cond, ColumnProxy):
        return cond.conds
    if not isinstance(cond, str_types):
        return cond
    m = COND_RE.match(cond)
    if m is None:
        msg = 'Could not parse condition from {0}'
        raise ValueError(msg.format(cond))
    c = tuple(map(stripper, m.groups()))
    return c


EMPTY_SLICE = slice(None)

def has_no_conds(key):
    """Determines if a key means that there are no conditions given."""
    if isinstance(key, ColumnProxy):
        return False
    return ((key is Ellipsis) or (key is None) or (key == EMPTY_SLICE) or
        (isinstance(key, Sized) and len(key) == 0))


class MetricProxy(object):
    """A proxy metric for nicer spelling of metrics in an execution context.
    Objects of this class are bound to an evaluator and will return a data frame
    when indexed. Index parameters are interpreted as query conditions.
    """

    def __init__(self, name, evaler=None):
        """Parameters
        ----------
        name : str
            The metric name.
        evaler : cymetric.evaluator.Evaluator, optional
            The evaluator for the metrics, required if db not given.
        """
        self.name = name
        self.evaler = evaler

    def __getitem__(self, key):
        """Evaluates if conditions are given for evaluation of metric and
        formats them.
        """
        if has_no_conds(key):
            conds = None
        elif isinstance(key, ColumnProxy):
            conds = key.conds
        elif isinstance(key, str_types):
            conds = [parse_cond(key)]
        else:
            conds = []
            for k in key:
                if isinstance(k, ColumnProxy):
                    conds += k.conds
                elif isinstance(k, str_types):
                    conds.append(parse_cond(k))
        return self.evaler.eval(self.name, conds=conds)


class ExecutionContext(MutableMapping):
    """An execution context for the command line or any other situation
    that involves the automatic injection of metric names, column names,
    etc.
    """

    def __init__(self, evaler=None, db=None, *args, **kwargs):
        """Parameters
        ----------
        evaler : cymetric.evaluator.Evaluator, optional
            The evaluator for the metrics, required if db not given.
        db : database, optional
            Required if evaler not given.
        args, kwargs : tuple, dict
            Other arguments to the mutable mapping.
        """
        self._ctx = ctx = {}
        self.evaler = evaler or evaluator.Evaluator(db)
        import cymetric as cym  # lazy import needed
        ctx['cym'] = cym
        ctx['np'] = np
        ctx['pd'] = pd
        ctx['uuid'] = uuid
        try:
            import matplotlib
            import matplotlib.pyplot as plt
            ctx['matplotlib'] = matplotlib
            ctx['plt'] = plt
        except ImportError:
            pass
        ctx.update(*args, **kwargs)

    def __getitem__(self, key):
        """Retrieves metric from registry or column from root metrics."""
        if key in __builtins__:
            raise KeyError
        elif key in METRIC_REGISTRY:
            self._ctx[key] = MetricProxy(key, self.evaler)
        elif key not in self._ctx:
            self._ctx[key] = ColumnProxy(key)
        return self._ctx[key]

    def __setitem__(self, key, value):
        """Sets metric with value indexed by key."""
        self._ctx[key] = value

    def __delitem__(self, key):
        """Deletes metric indexed by key."""
        del self._ctx[key]

    def __iter__(self):
        """Indexes metric by key."""
        for key in self._ctx:
            yield key

    def __len__(self):
        """Returns length of metric dict."""
        return len(self._ctx)

    def keys(self):
        """Returns copy of metric keys."""
        return self._ctx.keys()

    def values(self):
        """Returns copy of metric values."""
        return self._ctx.values()

    def items(self):
        """Returns copy of metrics in (key, value) form."""
        return self._ctx.items()

    def __del__(self):
        """Closes db, flushing remaining buffers."""
        self.evaler.db.flush()


def exec_code(code, db, write=True):
    """Runs a code snipper in the context of a database."""
    evaler = Evaluator(db, write=write)
    glb = {}
    loc = ExecutionContext(evaler=evaler)
    exec(code, glb, loc)
